#include <stdlib.h>
#include <errno.h>

#include "dequeue.h"

////////////////////////////////////////////////////////////////////////////////
struct dequeue * dequeue_new(size_t const max)
   {
   if (max < 1)
      {
      errno = EINVAL;
      return NULL;
      }

   struct dequeue * deq = malloc(sizeof(struct dequeue));
   if (NULL == deq)
      return NULL;

   deq->bottom = deq->top = deq->count = 0;
   deq->max = max;

   deq->data = calloc(deq->max, sizeof(void *));
   if (NULL == deq->data)
      {
      free(deq);
      return NULL;
      }

   return deq;
   }

////////////////////////////////////////////////////////////////////////////////
// Returns:
// NULL on error
// elem on success and the buffer is not full
// the overwritten element on success if the buffer was full
void * dequeue_push_top(struct dequeue * deq, void * elem)
   {
   if ((NULL == deq) || (NULL == elem))
      {
      errno = EINVAL;
      return NULL;
      }

   if (deq->count == deq->max)
      {
      errno = ENOMEM;
      return NULL;
      }

   if (deq->max - 1 == deq->top)
      deq->top = 0;
   else
      deq->top++;

   deq->data[deq->top] = elem;
   deq->count++;

   return elem;
   }

////////////////////////////////////////////////////////////////////////////////
void * dequeue_push_bottom(struct dequeue * deq, void * elem)
   {
   if ((NULL == deq) || (NULL == elem))
      {
      errno = EINVAL;
      return NULL;
      }

   if (deq->count == deq->max)
      {
      errno = ENOMEM;
      return NULL;
      }

   if (0 == deq->bottom)
      deq->bottom = deq->max - 1;
   else
      deq->bottom--;

   deq->data[deq->bottom] = elem;
   deq->count++;

   return elem;
   }

////////////////////////////////////////////////////////////////////////////////
void * dequeue_pop_top(struct dequeue * deq)
   {
   if (NULL == deq)
      {
      errno = EINVAL;
      return NULL;
      }

   if (deq->count == 0)
      return NULL;

   void * t = deq->data[deq->top];

   if (deq->top == 0)
      deq->top = deq->max - 1;
   else
      deq->top--;

   deq->count--;

   return t;
   }

////////////////////////////////////////////////////////////////////////////////
void * dequeue_pop_bottom(struct dequeue * deq)
   {
   if (NULL == deq)
      {
      errno = EINVAL;
      return NULL;
      }

   if (deq->count == 0)
      return NULL;

   void * t = deq->data[deq->bottom];

   if (deq->bottom == deq->max - 1)
      deq->bottom = 0;
   else
      deq->bottom++;

   deq->count--;

   return t;
   }


////////////////////////////////////////////////////////////////////////////////
size_t dequeue_size(struct dequeue * deq)
   {
   if (NULL == deq)
      {
      errno = EINVAL;
      return 0;
      }
   return deq->count;
   }

////////////////////////////////////////////////////////////////////////////////
void dequeue_delete(struct dequeue * deq)
   {
   if (NULL != deq)
      {
      free(deq->data);
      free(deq);
      }
   }
